// comets.h		ver.0.001
//copyright 2016 Beer[Tsutsumi Hiroyuki]

#include	<Be.h>

#if	B_BEOS_VERSION>=0x0520
#include	<Locale.h>
typedef	const char	Char;
#else
#define	_T(str)	str
typedef	char		Char;
#endif

#ifdef	B_HAIKU_VERSION
#define	MENUBAR_HEIGHT	19
#else
#define	MENUBAR_HEIGHT	20
#endif

union BGRA{				//bgra
	uint32	c;
	struct	{uchar	b,g,r,a;}u;
};

struct label{
	int		lno;	//line no.
	uint16	adr;
	char	name[9];
};

struct module{
	int		sline,eline;
	int		slbl,elbl;
	int		entl;			//start de shiteishita sakino label
};

struct line{
	char	lbl[9];
	char	opc[7];
	char	opr[240];
	char	rem[256];
};

struct opr{
	char	a[128];
	char	b[128];
	char	c[128];
};

enum flags{
	of=0x01,
	sf=0x02,
	zf=0x04,
};

enum chrset{
	sjis=0x01,
};

#define	MAXSRCSIZE	0x40000
#define	MAXSRCLINE	10000
#define	MAXLABEL	1000
#define	MAXMODULE	100

//graphic-ram		8000-bfff
//chr-gen(color)	c000-cfff
//chr-gen(mono) 	d000-d3ff
//text-ram 			d400-d7ff
//palette		 	d800-da3f
//control			da40-da41
//key				da50-da5f

class comet
{
	public:
		comet();
		~comet();
		void	init(bool *inpf,char *inp,int inpl,bool *outf,char *out,int outl,char *rp);//init sys
		void	reset(char flag);	//init cpu
		void	run(int steps);
		int32	sys_main(void *data);
		int32	cpu_main(void *data);
		int32	snd_main(void *data);
		void	setfr(int acc);
		void	setfr1(int acc);
		void	setfr2(int acc);
		void	setfr3(int acc,uint16 cy);
		uint16	pr,sp,fr;
		uint16	gr[16];		//8-15...ext
		uint16	mem[0x10000],psize;
		int		svc;		//set -1 when svc-return
		bool	svcw;		//wait svc-complete
		int		job;		//0.init	1.run  -1.end
		char	scmode;		//screen_mode
		int		cset;	//in/out character set
	public:
		char	flag;
		uint32	count;
	private:
		thread_id	t_cpu_id,t_sys_id;
		int		stp;
		bool	*inpf,*outf;
		char	*inp,*out;
		int		inpl,outl;
		char	*rp;
};

class caslassembler
{
	public:
		caslassembler();
		~caslassembler();
		void	init(char *rp,uint16 *mem,uint16 *psize);
		int		assemble(char *src,off_t ssize,char flag);	//init memory,encode,set binary
	private:
		int		encode1();	//step1  pickup label,set adr with label
		int		encode2();	//step2  encode with label   ret.code_bytes
		int		add_lbl(int mno,char *s,int lno,uint16 adr);
		int		search_lbl(int mno,char *s);
		int		split_line(int lno,struct line &l);
		int		split_opr(char *s,struct opr &opr);
		int		enc_adr(int mno,char *s,uint16 *d=NULL);
		int		enc_line(int mno,struct line &l,uint16 *d=NULL);
		int		macro(int mno,struct line &l,uint16 *d=NULL);	//encode macro  ret.bytes (err=-1)
		int		ds(char *s,uint16 *d=NULL);	//encode ds  ret.bytes (err=-1)
		int		dc(int mno,char *s,uint16 *d=NULL);	//encode dc  ret.bytes  mno=-1.literal
	public:
		char	*src;
		off_t	ssize;
		int		lc;		//total line
		int		*slp;	//offset(pointer each line to source)	-1:end
		struct label	lbl[MAXLABEL+1];
		int		lblc;
		struct module	pro[MAXMODULE+1];
		int		proc;	//number of program
		uint16	codeadr[MAXSRCLINE+1];
		char	err[256];	//errmes
	private:
		char	*rp;	//report
		uint16	*mem;
		uint16	*psize;	//program(binary) size
		char	flag;
};

class cometsview:public BView
{
	public:
		cometsview(BRect frame);
		~cometsview();
		void	AttachedToWindow();
		void	Draw(BRect rect);
		void	Pulse();
		void	MouseDown(BPoint pt);
//		void	MouseUp(BPoint pt);
//		void	MouseMoved(BPoint pt, uint32 code, const BMessage *msg);
		void	MessageReceived(BMessage *msg);
//		BMessage*	ddmessage();		//for drag&drop
//		void	ddcopy(BMessage *msg);	//for drag&drop
//		void	select_trans(const char *mime,translator_id *tid,uint32 *fmt);
//		void	redraw(bool rf=true);
//		void	drawframe(const s_sel &fr);
//		void	eraseframe(const s_sel &fr);
		void	load(entry_ref &eref);
		void	reset();
//		void	save(entry_ref &dir,char *name);
	public:
	private:
		BFilePanel	*fp;
		char		title[64];

		char	*src;
		off_t	ssize;
//		caslassembler	cas;
//		comet			cpu;
		int		job;		//0.init  1.load  2.asm  3.stop  4.run  5.svc  6.end
		int		cset;	//in/out character set
		char	flag;		//checkbox  0.asm_report 1.pr_check 2.sp_check 3.ext_mode
		uint16	psize;

		BButton	*b_reset,*b_start,*b_stop,*b_step;
		BCheckBox	*cb_prc,*cb_spc,*cb_opc,*cb_svc,*cb_rpt,*cb_exc,*cb_exh;
	public:		//for terminal
		char	input[256],output[8192];
		bool	inpf,outf;
		char	report[1024];	//from asm/cpu
};

class termview:public BTextView
{
	public:
		termview(BRect frame,BRect textRect);
		~termview();
		void	AttachedToWindow();
//		void	Draw(BRect rect);
		void	Pulse();
		void	KeyDown(const char* bytes,int32 numBytes);
//	virtual	void				MessageReceived(BMessage* message);
		void	puts(char *s);
	public:
//		int		job=0;		//0..none 1..input 2..inputend
		char	instr[256];
		int		inlen;
};

class scrnview:public BView
{
	public:
		scrnview(BRect frame);
void test();
//		~scrnview();
//		void	AttachedToWindow();
		void	Draw(BRect rect);
		void	Pulse();
//		void	KeyDown(const char* bytes,int32 numBytes);
//	virtual	void				MessageReceived(BMessage* message);
		void	overlap(uint32 *p,uint32 c);
		void	loadfont();
	private:
		char	*font;
		uint16	*gram,*cgc,*cgm,*tram,*pal,*ctrl,*key;	//gram,cg1,cg2,tram,pal,ctrl,key
		BBitmap	*vbm,*bm;
		uint32	count;	//if(count!=cpu->count)redraw
};

class cometswindow:public BWindow
{
	public:
		cometswindow(BRect frame);
		bool	QuitRequested();
//		void	FrameResized(float width,float height);
//		void	resizeview(int width,int height);
		void	setupmenus();
//		void	get_translator_for_output();
		cometsview	*v;
//		termview	*tv;	//terminal(interface)
		BTextView	*sv;	//status
		BTextView	*rv;	//register
		BScrollBar	*sbv,*sbh;
};

class termwindow:public BWindow
{
	public:
		termwindow(BRect frame);
//		bool	QuitRequested();
		termview	*tv;	//terminal(interface)
};

class scrnwindow:public BWindow
{
	public:
		scrnwindow(BRect frame);
//		bool	QuitRequested();
		scrnview	*sv;	//screen(ext_mode)
};

class comets:public BApplication
{
	public:
		comets(int argc,char **argv);
		void	RefsReceived(BMessage *mes);
//		void	MessageReceived(BMessage *mes);
		void	AboutRequested();
		cometswindow	*w;
		termwindow		*tw;
		scrnwindow		*sw;
};

caslassembler	*cas;
comet			*cpu;

int32 thread_sys(void *data)
{
	cpu->sys_main(data);
}
int32 thread_cpu(void *data)
{
	cpu->cpu_main(data);
}
int32 thread_snd(void *data)
{
	cpu->snd_main(data);
}

const int16	opck1[256]={	//opcode check (normal)
	0xff,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,0xf8,0x8f,  -1,  -1,  -1,  -1,0xf8,
	  -1,0xf8,  -1,  -1,  -1,0xf8,  -1,0xf8,  -1,0xf8,  -1,0xf8,  -1,0xf8,  -1,  -1,
	  -1,0xf8,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,
	0xff,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,
	  -1,0x88,  -1,0x88,  -1,0x88,  -1,0x88,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,
	  -1,0x88,  -1,0x88,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,
	0x88,0x88,0x88,0x88,0x88,0x88,0x88,0x88,  -1,  -1,  -1,  -1,  -1,  -1,  -1,  -1,
	0x88,0x88,0x88,0x88,  -1,  -1,  -1,  -1,0x88,0x88,0x88,0x88,0x88,0x88,0x88,0x88
};

const uint16	cnv[512]={	//1bytecode>sjis (ascii/katakana,greek/hiragana)
	0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
	0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
	0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
	0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
	0x8140,0x8149,0x8168,0x8194,0x8190,0x8193,0x8195,0x8166,
	0x8169,0x816a,0x8196,0x817b,0x8143,0x817c,0x8144,0x815e,
	0x824f,0x8250,0x8251,0x8252,0x8253,0x8254,0x8255,0x8256,
	0x8257,0x8258,0x8146,0x8147,0x8183,0x8181,0x8184,0x8148,
	0x8197,0x8260,0x8261,0x8262,0x8263,0x8264,0x8265,0x8266,
	0x8267,0x8268,0x8269,0x826a,0x826b,0x826c,0x826d,0x826e,
	0x826f,0x8270,0x8271,0x8272,0x8273,0x8274,0x8275,0x8276,
	0x8277,0x8278,0x8279,0x816d,0x818f,0x816e,0x814f,0x8151,
	0x814d,0x8281,0x8282,0x8283,0x8284,0x8285,0x8286,0x8287,
	0x8288,0x8289,0x828a,0x828b,0x828c,0x828d,0x828e,0x828f,
	0x8290,0x8291,0x8292,0x8293,0x8294,0x8295,0x8296,0x8297,
	0x8298,0x8299,0x829a,0x816f,0x8162,0x8170,0x8150,0x0000,
	0x849f,0x84a0,0x84a1,0x84a2,0x84a3,0x84a4,0x84a5,0x84a6,
	0x84a7,0x84a8,0x84a9,0x84aa,0x84ab,0x84ac,0x84ad,0x84ae,
	0x84af,0x84b0,0x84b1,0x84b2,0x84b3,0x84b4,0x84b5,0x84b6,
	0x84b7,0x84b8,0x84b9,0x84ba,0x84bb,0x84bc,0x84bd,0x84be,
	0x0000,0x8142,0x8175,0x8176,0x8141,0x8145,0x8392,0x8340,
	0x8342,0x8344,0x8346,0x8348,0x8383,0x8385,0x8387,0x8362,
	0x815b,0x8341,0x8343,0x8345,0x8347,0x8349,0x834a,0x834c,
	0x834e,0x8350,0x8352,0x8354,0x8356,0x8358,0x835a,0x835c,
	0x835e,0x8360,0x8363,0x8365,0x8367,0x8369,0x836a,0x836b,
	0x836c,0x836d,0x836e,0x8371,0x8374,0x8377,0x837a,0x837d,
	0x837e,0x8380,0x8381,0x8382,0x8384,0x8386,0x8388,0x8389,
	0x838a,0x838b,0x838c,0x838d,0x838f,0x8393,0x814a,0x814b,
	0x8395,0x8396,0x838e,0x8394,0x8390,0x8391,0x834b,0x834d,
	0x834f,0x8351,0x8353,0x8355,0x8357,0x8359,0x835b,0x835d,
	0x835f,0x8361,0x8364,0x8366,0x8368,0x8370,0x8373,0x8376,
	0x8379,0x837c,0x836f,0x8372,0x8375,0x8378,0x837b,0x0000,

	0x0000,0x81f4,0x81f2,0x81f3,0x81a8,0x81a9,0x81aa,0x81ab,
	0x81b8,0x81b9,0x81ba,0x81bb,0x81bc,0x81bd,0x81be,0x81bf,
	0x819e,0x819f,0x81a0,0x81a1,0x81a2,0x81a3,0x81a4,0x81a5,
	0x8199,0x819a,0x819b,0x819c,0x819d,0x81a6,0x8189,0x818a,
	0x8140,0x8149,0x81da,0x81db,0x8191,0x81f1,0x8192,0x81df,
	0x8780,0x8781,0x817e,0x817d,0x8185,0x8182,0x8186,0x8180,
	0x824f,0x8740,0x8741,0x8742,0x8743,0x8744,0x8745,0x8746,
	0x8747,0x8748,0x8187,0x81e3,0x81e1,0x81e0,0x81e2,0x8148,
	0x81dd,0x839f,0x83a0,0x83a1,0x83a2,0x83a3,0x83a4,0x83a5,
	0x83a6,0x83a7,0x83a8,0x83a9,0x83aa,0x83ab,0x83ac,0x83ad,
	0x83ae,0x83af,0x83b0,0x83b1,0x83b2,0x83b3,0x83b4,0x83b5,
	0x83b6,0x0000,0x0000,0x816b,0x815f,0x816c,0x8188,0x81e6,
	0x8198,0x83bf,0x83c0,0x83c1,0x83c2,0x83c3,0x83c4,0x83c5,
	0x83c6,0x83c7,0x83c8,0x83c9,0x83ca,0x83cb,0x83cc,0x83cd,
	0x83ce,0x83cf,0x83d0,0x83d1,0x83d2,0x83d3,0x83d4,0x83d5,
	0x83d6,0x0000,0x0000,0x8179,0x8161,0x817a,0x8160,0x0000,
	0x81c8,0x81c9,0x81e7,0x81e8,0x81f5,0x81f6,0x8793,0x81f7,
	0x81de,0x81dc,0x818e,0x81f0,0x81cb,0x81cc,0x81ac,0x81a7,
	0x8152,0x8153,0x8154,0x8155,0x8156,0x8157,0x8158,0x8159,
	0x8165,0x8166,0x8167,0x8168,0x8177,0x8178,0x8163,0x8164,
	0x0000,0x8142,0x8175,0x8176,0x8141,0x8145,0x82f0,0x829f,
	0x82a1,0x82a3,0x82a5,0x82a7,0x82e1,0x82e3,0x82e5,0x82c1,
	0x815b,0x82a0,0x82a2,0x82a4,0x82a6,0x82a8,0x82a9,0x82ab,
	0x82ad,0x82af,0x82b1,0x82b3,0x82b5,0x82b7,0x82b9,0x82bb,
	0x82bd,0x82bf,0x82c2,0x82c4,0x82c6,0x82c8,0x82c9,0x82ca,
	0x82cb,0x82cc,0x82cd,0x82d0,0x82d3,0x82d6,0x82d9,0x82dc,
	0x82dd,0x82de,0x82df,0x82e0,0x82e2,0x82e4,0x82e6,0x82e7,
	0x82e8,0x82e9,0x82ea,0x82eb,0x82ed,0x82f1,0x814a,0x814b,
	0x0000,0x0000,0x82ec,0x0000,0x82ee,0x82ef,0x82aa,0x82ac,
	0x82ae,0x82b0,0x82b2,0x82b4,0x82b6,0x82b8,0x82ba,0x82bc,
	0x82be,0x82c0,0x82c3,0x82c5,0x82c7,0x82cf,0x82d2,0x82d5,
	0x82d8,0x82db,0x82ce,0x82d1,0x82d4,0x82d7,0x82da,0x0000,
};

const uchar	xfd[4][64]={
	 0, 1, 2, 3, 4, 5, 6, 7,
	 8, 9,10,11,12,13,14,15,
	16,17,18,19,20,21,22,23,
	24,25,26,27,28,29,30,31,
	32,33,34,35,36,37,38,39,
	40,41,42,43,44,45,46,47,
	48,49,50,51,52,53,54,55,
	56,57,58,59,60,61,62,63,
	56,48,40,32,24,16, 8, 0,
	57,49,41,33,25,17, 9, 1,
	58,50,42,34,26,18,10, 2,
	59,51,43,35,27,19,11, 3,
	60,52,44,36,28,20,12, 4,
	61,53,45,37,29,21,13, 5,
	62,54,46,38,30,22,14, 6,
	63,55,47,39,31,23,15, 7,
	63,62,61,60,59,58,57,56,
	55,54,53,52,51,50,49,48,
	47,46,45,44,43,42,41,40,
	39,38,37,36,35,34,33,32,
	31,30,29,28,27,26,25,24,
	23,22,21,20,19,18,17,16,
	15,14,13,12,11,10, 9, 8,
	 7, 6, 5, 4, 3, 2, 1, 0,
	 7,15,23,31,39,47,55,63,
	 6,14,22,30,38,46,54,62,
	 5,13,21,29,37,45,53,61,
	 4,12,20,28,36,44,52,60,
	 3,11,19,27,35,43,51,59,
	 2,10,18,26,34,42,50,58,
	 1, 9,17,25,33,41,49,57,
	 0, 8,16,24,32,40,48,56
};
